#include "mainwindow.h"
#include "odlanrfid_demo_global.h"
#include "preferences.h"

#include <ui_mainwindow.h>
#include <QMessageBox>

#include <QStandardPaths>
#include <QNetworkDatagram>
#include <QDir>
#include <QDebug>
#include <QBuffer>
#include <QDateTime>

#include <QSqlQuery>
#include <QSqlError>

MainWindow::MainWindow(QWidget *parent) :
    QMainWindow(parent),
    ui(new Ui::MainWindow),
    run_mode(Scanning),
    set(this), socket(this), cleanTimer(this)
{
    ui->setupUi(this);
    connect(ui->actionExit, SIGNAL(triggered(bool)), qApp, SLOT(quit()));
    loaderGif = new QMovie(":/images/loader.gif", "GIF", this);
    cleanTimer.setInterval(3000);
    connect(&cleanTimer, SIGNAL(timeout()), this, SLOT(clean_timeout()));
    connect(&socket, SIGNAL(readyRead()), this, SLOT(socket_readyRead()));
    clean_timeout();
}

MainWindow::~MainWindow()
{
    delete ui;
}

bool MainWindow::readSettings()
{
    if(!set.contains(SETTINGS_PORT))
    {
        //assume this is the first run
        on_actionPreferences_triggered();
        if(!set.contains(SETTINGS_PORT))
        {
            //cancelled
            qWarning()<<"No valid settings";
            return false;
        }
    }

    QSqlDatabase db = QSqlDatabase::addDatabase("QSQLITE", DB_CONNECTION);
    QString defaultDataDir = QStandardPaths::standardLocations(QStandardPaths::AppDataLocation).at(0);
    if(!QDir(defaultDataDir).exists())
    {
        QDir().mkpath(defaultDataDir);
    }
    db.setDatabaseName(defaultDataDir+"/odlanrfiddemo.db");
    if(!db.open())
    {
        qWarning()<<"Database cannot be opened";
        return false;
    }
    initDatabase();

    return initSocket();
}

void MainWindow::socket_readyRead()
{
    while (socket.hasPendingDatagrams())
    {
        QNetworkDatagram datagram = socket.receiveDatagram();
        processDatagram(datagram);
    }
}

void MainWindow::clean_timeout()
{
    ui->avatarLabel->setReadOnly(true);
    ui->infoView->setReadOnly(true);
    ui->idField->setText(tr("Scanning..."));
    ui->validField->setDateTime(QDateTime::currentDateTime());
    ui->infoView->clear();
    ui->avatarLabel->setMovie(loaderGif);
    loaderGif->start();
}

void MainWindow::on_actionPreferences_triggered()
{
    Preferences prefsDlg(this);
    prefsDlg.readSettings();
    if(prefsDlg.exec() == QDialog::Accepted)
    {
        prefsDlg.writeSettings();
    }
}

void MainWindow::on_actionAdd_Entry_triggered()
{
    cleanTimer.stop();
    loaderGif->stop();
    ui->idField->clear();
    ui->validField->setReadOnly(false);
    ui->infoView->clear();
    ui->infoView->setReadOnly(false);
    ui->avatarLabel->setReadOnly(false);
    ui->avatarLabel->setText(tr("Drop an image here"));
    run_mode = NewEntry;
}

void MainWindow::on_actionAbout_triggered()
{
    QMessageBox::about(this,tr("ODLAN-RFID Demo"), tr("Hello World!"));
}

void MainWindow::on_actionClean_Database_triggered()
{
    if(QMessageBox::question(this, tr("Clean Database"), tr("Are you sure? This <b>cannot</b> be undone!"),
                             QMessageBox::Yes|QMessageBox::No, QMessageBox::No) == QMessageBox::Yes) {
        QSqlQuery query(QSqlDatabase::database(DB_CONNECTION));
        query.exec("DELETE * FROM ODLANRFID");
    }
}

void MainWindow::initDatabase()
{
    QSqlQuery query(QSqlDatabase::database(DB_CONNECTION));

    if(!query.exec("SELECT VERSION FROM BUILDINFO WHERE NAME='ODLANRFID'") || !query.next())
    {
        //no buildinfo -> invalid table;
        qDebug()<<"Empty/Invalid database - no buildinfo";
        query.exec("DROP TABLE IF EXISTS BUILDINFO");
        query.exec("DROP TABLE IF EXISTS ODLANRFID");
    }
    else
    {
        QString ver = query.value(0).toString();
        qDebug()<<"Database version "<<ver;
        if((ver.split(".",QString::SkipEmptyParts).size()!=2)||
           (ver.split(".",QString::SkipEmptyParts).first().toInt()!=TABLE_VERSION_MAJ)||
           (ver.split(".",QString::SkipEmptyParts).last().toInt()!=TABLE_VERSION_MIN))
        {
            qWarning()<<"Database is outdated "<<ver<<" vs "
                     <<TABLE_VERSION_MAJ<<"."<<TABLE_VERSION_MIN<<" -> dropping";
            query.exec("DROP TABLE IF EXISTS BUILDINFO");
            query.exec("DROP TABLE IF EXISTS MIFARECLASSIC");
        }
    }
    query.finish();

    QSqlDatabase::database(DB_CONNECTION).transaction();
    if(!query.exec("CREATE TABLE IF NOT EXISTS BUILDINFO(NAME TEXT UNIQUE, DATE TEXT, VERSION TEXT)"))
    {
        qWarning()<<"database: buildinfo create failure - "<<query.lastError().text();
        return;
    }

    query.prepare("INSERT OR IGNORE INTO BUILDINFO VALUES(?,?,?)");
    query.addBindValue("ODLANRFID");
    query.addBindValue(QDateTime::currentDateTimeUtc().toString("yyyy.MM.dd hh:mm:ss"));
    query.addBindValue(QString("%1.%2").arg(TABLE_VERSION_MAJ).arg(TABLE_VERSION_MIN));
    if(!query.exec())
    {
        qWarning()<<"database: buildinfo query execution failure - "<<query.lastError().text();
        return;
    }
    query.finish();

    if(!query.exec("CREATE TABLE IF NOT EXISTS ODLANRFID(UID TEXT NOT NULL, VALID INTEGER, AVATAR BLOB, INFO TEXT)"))
    {
        qWarning()<<"database: table creation failure - "<<query.lastError().text();
        return;
    }
    query.finish();
    QSqlDatabase::database(DB_CONNECTION).commit();
}

bool MainWindow::initSocket()
{
    bool res = socket.bind(set.value(SETTINGS_PORT).toUInt(), QUdpSocket::ReuseAddressHint);
    qDebug()<<"Bind to "<<set.value(SETTINGS_PORT).toUInt()<<" port: "<<(res ? "success" : "failure");
    return res;
}

void MainWindow::displayInvalid(const QString& uuid)
{
    ui->idField->setText(uuid);
    ui->validField->clear();
    ui->infoView->setText(tr("Unknown ID!"));
    loaderGif->stop();
    ui->avatarLabel->setPixmap(QPixmap(":/images/dialog-warning.png").scaled(ui->avatarLabel->size(),Qt::KeepAspectRatio, Qt::SmoothTransformation));
    cleanTimer.start();
}

bool MainWindow::displayUID(const QByteArray& id)
{
    cleanTimer.stop();
    QString uuid = QString::fromLatin1(id.toUpper());
    if(ui->idField->text() == uuid) {
        qWarning()<<"Duplicate - ignore";
        return true;
    }

    QSqlQuery query(QSqlDatabase::database(DB_CONNECTION));
    qDebug()<<"Search "<<uuid<<" in database";
    if(query.exec(QString("SELECT VALID,AVATAR,INFO FROM ODLANRFID WHERE UID='%1'").arg(uuid)) && query.next())
    {
        ui->idField->setText(uuid);
        QDateTime dt = QDateTime::fromSecsSinceEpoch(query.value(0).toUInt(),Qt::UTC).toLocalTime();
        if(!dt.isValid())
        {
            qWarning()<<"Datetime "<<query.value(0).toString()<<" is invalid";
            displayInvalid(uuid);
            return false;
        }
        ui->validField->setDateTime(dt);
        loaderGif->stop();
        ui->avatarLabel->setImage(query.value(1).toByteArray(),"PNG");
        ui->infoView->setText(query.value(2).toString());
        cleanTimer.start();
        return true;
    }

    qWarning()<<"Query"<<query.lastQuery()<<" failed: "<<query.lastError().text();
    displayInvalid(uuid);
    return false;
}

bool MainWindow::saveUID(const QByteArray& id)
{
    cleanTimer.stop();
    QString uuid = QString::fromLatin1(id.toUpper());
    QSqlQuery query(QSqlDatabase::database(DB_CONNECTION));
    if((ui->validField->dateTime() < QDateTime::currentDateTime()) || !ui->avatarLabel->hasImage())
    {
        QMessageBox::warning(this, tr("New Entry"), tr("'Valid Through' value and an 'Avatar' are mandatory."));
        return false;
    }

    if(query.exec(QString("SELECT VALID FROM ODLANRFID WHERE UID='%1'").arg(uuid)) && query.next())
    {
        if(QMessageBox::warning(this, tr("New Entry"), tr("An entry with UID %1 already exists. Update?"),
                                         QMessageBox::Yes|QMessageBox::No, QMessageBox::No) != QMessageBox::Yes)
        {
            return false;
        }
    }

    query.prepare(QString("INSERT OR REPLACE INTO ODLANRFID (UID, VALID, AVATAR, INFO) values (?, ?, ?, ?)"));
    query.addBindValue(uuid);
    query.addBindValue(ui->validField->dateTime().toUTC().toSecsSinceEpoch());
    query.addBindValue(ui->avatarLabel->getImage("PNG"));
    query.addBindValue(ui->infoView->toPlainText());
    if(!query.exec())
    {
        QMessageBox::warning(this, tr("New Entry"), tr("Cannot add new entry: %1").arg(query.lastError().text()));
        return false;
    }

    QMessageBox::information(this, tr("New Entry"), tr("%1 saved").arg(uuid));
    clean_timeout();
    return true;
}

void MainWindow::processDatagram(const QNetworkDatagram& datagram)
{
    /*
     * no ack:
     *  uid or block in hex + NL
     * ack:
     *  seq (1byte), uid or block in hex + NL
     *  reply: seq + 0
     */
    QByteArray id;
    char seq[2] = {0,0};
    bool ack = set.value(SETTINGS_ACK, false).toBool();

    if(ack)
    {
        if(datagram.data().size()> 1)
        {
            seq[0] = datagram.data()[0];
            id = datagram.data().mid(1).trimmed();
        }
    }
    else
    {
        id = datagram.data().trimmed();
    }

    if(id.isEmpty())
    {
        qWarning()<<"Invalid datagram: empty UID";
        return;
    }

    switch(run_mode)
    {
        case Scanning:
        {
            if(displayUID(id))
            {
                if(ack)
                {
                    socket.writeDatagram(QByteArray(seq,2), datagram.senderAddress(), datagram.senderPort());
                }
            }
            break;
        }
        case NewEntry:
        {
            saveUID(id);
            finishEditing();
            break;
        }
    }
}

void MainWindow::finishEditing()
{
    run_mode = Scanning;
    clean_timeout();
}
